/*
 * EDG, a library to generate and slice Expression Dependence Graphs.
 * Copyright (c) 2021. David Insa, Sergio Pérez, Josep Silva, Salvador Tamarit.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package edg.constraint;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.util.Set;

import edg.constraint.AccessConstraint.Operation;
import edg.slicing.Phase;

public class ConstraintFactory
{
	private static ConstraintFactory factory = new ConstraintFactory();
	public static ConstraintFactory getFactory()
	{
		return ConstraintFactory.factory;
	}

	private enum Type { DataConstructor, ListComprehension, List, AddNode, Asterisk, Empty, Phase }
	private Map<Type, Set<Constraint>> constraints = new Hashtable<Type, Set<Constraint>>();

	private ConstraintFactory()
	{
		for (Type type : Type.values())
			this.constraints.put(type, new HashSet<Constraint>());
		this.constraints.get(Type.Asterisk).add(AsteriskConstraint.getConstraint());
		this.constraints.get(Type.Empty).add(EmptyConstraint.getConstraint());
	}

	private Constraint getConstraint(Type type, Constraint constraint)
	{
		final Set<Constraint> constraints = this.constraints.get(type);

		for (Constraint oldConstraint : constraints)
			if (oldConstraint.equals(constraint))
				return oldConstraint;

		constraints.add(constraint);
		return constraint;
	}

	public DataConstructorConstraint getDataConstructorConstraint(Operation operation, String index)
	{
		final DataConstructorConstraint constraint = new DataConstructorConstraint(operation, index);
		return (DataConstructorConstraint) this.getConstraint(Type.DataConstructor, constraint);
	}
	public ListComprehensionConstraint getListComprehensionConstraint(Operation operation)
	{
		final ListComprehensionConstraint constraint = new ListComprehensionConstraint(operation);
		return (ListComprehensionConstraint) this.getConstraint(Type.ListComprehension, constraint);
	}
	public ListConstraint getListConstraint(Operation operation)
	{
		final ListConstraint constraint = new ListConstraint(operation);
		return (ListConstraint) this.getConstraint(Type.List, constraint);
	}
	public AddNodeConstraint getAddNodeConstraint(NodeConstraint nodeConstraint)
	{
		final AddNodeConstraint constraint = new AddNodeConstraint(nodeConstraint);
		return (AddNodeConstraint) this.getConstraint(Type.AddNode, constraint);
	}
	public AsteriskConstraint getAsteriskConstraint()
	{
		return (AsteriskConstraint) this.constraints.get(Type.Asterisk).iterator().next();
	}
	public EmptyConstraint getEmptyConstraint()
	{
		return (EmptyConstraint) this.constraints.get(Type.Empty).iterator().next();
	}
	public PhaseConstraint getPhaseConstraint(Phase... phases)
	{
		final PhaseConstraint constraint = new PhaseConstraint(phases);
		return (PhaseConstraint) this.getConstraint(Type.Phase, constraint);
	}
}