/*
 * EDG, a library to generate and slice Expression Dependence Graphs.
 * Copyright (c) 2021. David Insa, Sergio Pérez, Josep Silva, Salvador Tamarit.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package edg.constraint;

import java.util.LinkedList;
import java.util.List;

import edg.config.Config;
import edg.graph.Edge;
import edg.graph.Grammar;
import edg.graph.Node;
import edg.slicing.Phase;

public class GrammarConstraint extends EdgeConstraint
{
	private final Config config = Config.getConfig();
	private final Grammar grammar;
	private final Node refNode;

	public GrammarConstraint(Grammar grammar, Node refNode)
	{
		this.grammar = grammar;
		this.refNode = refNode;
	}

	public Node getRefNode()
	{
		return this.refNode;
	}

	public boolean equals(Object object)
	{
		if (object == this)
			return true;
		if (!(object instanceof GrammarConstraint))
			return false;

		final GrammarConstraint constraint = (GrammarConstraint) object;

		if (this.grammar != constraint.grammar)
			return false;
		if (this.refNode != constraint.refNode)
			return false;
		return true;
	}
	public String toString()
	{
		final String refNodeId = this.refNode.getData().getId() + "";
		return refNodeId;
// TODO Arreglar el getHashCode del WorkList antes de descomentar esto
/*
		final List<Constraints> productions = this.grammar.getProductions(this);
		if (productions.isEmpty())
			return new EmptyConstraint().toString();
		if (productions.size() != 1)
			return refNodeId;

		final Constraints production = productions.get(0);
		if (production.isEmpty())
			return new EmptyConstraint().toString();
		if (production.size() != 1)
			return refNodeId;

		final Constraint constraint = production.get(0);
		return constraint.toString();
*/
	}

	protected List<Constraints> resolve(Phase phase, Edge edge, Constraints constraints, int productionDepth)
	{
		if (phase.isInstanceof(Phase.Slicing))
			return this.resolveProductions(phase, edge, constraints, productionDepth);
		super.check(phase, Phase.SummaryGeneration);
		return super.wrap(super.push(phase, constraints));
	}
	protected List<Constraints> resolve(Phase phase, Edge edge, Constraints constraints, AccessConstraint topConstraint, int productionDepth)
	{
		if (phase.isInstanceof(Phase.Slicing))
			return this.resolveProductions(phase, edge, constraints, productionDepth);
		super.check(phase, Phase.SummaryGeneration);
		return super.wrap(super.push(phase, constraints));
	}
	protected List<Constraints> resolve(Phase phase, Edge edge, Constraints constraints, GrammarConstraint topConstraint, int productionDepth)
	{
		super.check(phase, Phase.SummaryGeneration);
		return super.wrap(super.push(phase, constraints));
	}
	protected List<Constraints> resolve(Phase phase, Edge edge, Constraints constraints, SeekingConstraint topConstraint, int productionDepth)
	{
		if (topConstraint.operation == SeekingConstraint.Operation.Add)
			if (phase.isInstanceof(Phase.Slicing))
				return this.resolveProductions(phase, edge, constraints, productionDepth);

		super.check(phase, Phase.SummaryGeneration);
		return super.wrap(super.push(phase, constraints));
	}
	protected List<Constraints> resolve(Phase phase, Edge edge, Constraints constraints, AsteriskConstraint topConstraint, int productionDepth)
	{
		super.check(phase, Phase.SummaryGeneration);
		return super.wrap(super.push(phase, constraints));
	}

	private List<Constraints> resolveProductions(Phase phase, Edge edge, Constraints constraints, int productionDepth)
	{
		super.check(phase, Phase.Slicing);
		if (productionDepth == this.config.maxProductionDepth)
			throw new StackOverflowError();

		final List<Constraints> newConstraintsList = new LinkedList<Constraints>();
		final List<Constraints> productions = this.grammar.getProductions(this);

		for (Constraints production : productions)
		{
			final Constraints constraintsClone = (Constraints) constraints.clone();
			final List<Constraints> pendingConstraintsList = new LinkedList<Constraints>();
			pendingConstraintsList.add(constraintsClone);
			final int productionSize = production.sizeEdgeConstraints();

			for (int constraintIndex = 0; constraintIndex < productionSize; constraintIndex++)
			{
				final EdgeConstraint constraint = production.getEdgeConstraint(constraintIndex);
				final List<Constraints> resolvedConstraintsList = new LinkedList<Constraints>();

				for (Constraints pendingConstraints : pendingConstraintsList)
				{
					final EdgeConstraint topConstraint = pendingConstraints.isEdgeConstraintsEmpty() ? null : pendingConstraints.peekEdgeConstraint();
					resolvedConstraintsList.addAll(constraint.resolve(phase, edge, pendingConstraints, topConstraint, productionDepth + 1));
				}
				pendingConstraintsList.clear();
				pendingConstraintsList.addAll(resolvedConstraintsList);
			}

			newConstraintsList.addAll(pendingConstraintsList);
		}

		return newConstraintsList;
	}
}